# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import netrc
import requests
import socket
import json
import struct
import ctypes
import time
import logging
import warnings
import numpy
import geopandas
from shapely.geometry import Polygon
from shapely.geometry.multipolygon import MultiPolygon
from datetime import datetime
from sliderule import version

###############################################################################
# GLOBALS
###############################################################################

# Coordinate Reference system definition for WGS84 Ellipsoid ensemble
# Coordinate system on the surface of a sphere or ellipsoid of reference.
EPSG_WGS84 = "EPSG:4326"

# This is 3D CRS for ITRF2014 realization with 2010.0 epoch: https://epsg.io/7912
# Note: using GRS80 ellipsoid, negligible difference in inverse flattening from WGS84
SLIDERULE_EPSG = "EPSG:7912"

PUBLIC_URL = "slideruleearth.io"
PUBLIC_ORG = "sliderule"

DEFAULT_TRUST_ENV = False

service_url = PUBLIC_URL
service_org = PUBLIC_ORG

session = requests.Session()
session.trust_env = DEFAULT_TRUST_ENV

ps_refresh_token = None
ps_access_token = None
ps_token_exp = None

MAX_PS_CLUSTER_WAIT_SECS = 600

request_timeout = (10, 120) # (connection, read) in seconds
decode_aux = True

logger = logging.getLogger(__name__)
console = None

rethrow_exceptions = False

clustering_enabled = False
try:
    from sklearn.cluster import KMeans
    clustering_enabled = True
except:
    logger.info("Unable to import sklearn... clustering support disabled")

recdef_table = {}

arrow_file_table = {}

profiles = {}

gps_epoch = datetime(1980, 1, 6)
tai_epoch = datetime(1970, 1, 1, 0, 0, 10)

eventformats = {
    "TEXT":     0,
    "JSON":     1
}

eventlogger = {
    0: logger.debug,
    1: logger.info,
    2: logger.warning,
    3: logger.error,
    4: logger.critical
}

exceptioncodes = {
    "INFO": 0,
    "ERROR": -1,
    "TIMEOUT": -2,
    "RESOURCE_DOES_NOT_EXIST": -3,
    "EMPTY_SUBSET": -4,
    "SIMPLIFY": -5
}

datatypes = {
    "TEXT":     0,
    "REAL":     1,
    "INTEGER":  2,
    "DYNAMIC":  3
}

basictypes = {
    "INT8":     { "fmt": 'b', "size": 1, "nptype": numpy.int8   },
    "INT16":    { "fmt": 'h', "size": 2, "nptype": numpy.int16  },
    "INT32":    { "fmt": 'i', "size": 4, "nptype": numpy.int32  },
    "INT64":    { "fmt": 'q', "size": 8, "nptype": numpy.int64  },
    "UINT8":    { "fmt": 'B', "size": 1, "nptype": numpy.uint8  },
    "UINT16":   { "fmt": 'H', "size": 2, "nptype": numpy.uint16 },
    "UINT32":   { "fmt": 'I', "size": 4, "nptype": numpy.uint32 },
    "UINT64":   { "fmt": 'Q', "size": 8, "nptype": numpy.uint64 },
    "BITFIELD": { "fmt": 'x', "size": 0, "nptype": numpy.byte   },  # unsupported
    "FLOAT":    { "fmt": 'f', "size": 4, "nptype": numpy.single },
    "DOUBLE":   { "fmt": 'd', "size": 8, "nptype": numpy.double },
    "TIME8":    { "fmt": 'q', "size": 8, "nptype": numpy.int64  }, # numpy.datetime64
    "STRING":   { "fmt": 's', "size": 1, "nptype": numpy.byte   }
}

codedtype2str = {
    0:  "INT8",
    1:  "INT16",
    2:  "INT32",
    3:  "INT64",
    4:  "UINT8",
    5:  "UINT16",
    6:  "UINT32",
    7:  "UINT64",
    8:  "BITFIELD",
    9:  "FLOAT",
    10: "DOUBLE",
    11: "TIME8",
    12: "STRING"
}

###############################################################################
# CLIENT EXCEPTIONS
###############################################################################

class FatalError(RuntimeError):
    pass

class TransientError(RuntimeError):
    pass

class RetryRequest(RuntimeError):
    pass

###############################################################################
# UTILITIES
###############################################################################

#
# __StreamSource
#
class __StreamSource:
    def __init__(self, data):
        self.source = data
    def __iter__(self):
        for line in self.source.iter_content(None):
            yield line

#
# __BufferSource
#
class __BufferSource:
    def __init__(self, data):
        self.source = data
    def __iter__(self):
        yield self.source

#
#  __populate
#
def __populate(rectype):
    global recdef_table
    if rectype not in recdef_table:
        recdef_table[rectype] = source("definition", {"rectype" : rectype})
    return recdef_table[rectype]

#
#  __parse_json
#
def __parse_json(data):
    """
    data: request response
    """
    lines = []
    for line in data:
        lines.append(line)
    response = b''.join(lines)
    return json.loads(response)

#
#  __decode_native
#
def __decode_native(rectype, rawdata):
    """
    rectype: record type supplied in response (string)
    rawdata: payload supplied in response (byte array)
    """
    global recdef_table

    # initialize record
    rec = { "__rectype": rectype }

    # get/populate record definition #
    recdef = __populate(rectype)

    # iterate through each field in definition
    for fieldname in recdef.keys():

        # double underline (__) as prefix indicates meta data
        if fieldname.find("__") == 0:
            continue

        # get field properties
        field = recdef[fieldname]
        ftype = field["type"]
        offset = int(field["offset"] / 8)
        elems = field["elements"]
        flags = field["flags"]

        # do not process pointers
        if "PTR" in flags:
            continue

        # check for mvp flag
        if not decode_aux and "AUX" in flags:
            continue

        # get endianness
        if "LE" in flags:
            endian = '<'
        else:
            endian = '>'

        # decode basic type
        if ftype in basictypes:

            # check if array
            is_array = not (elems == 1)

            # get number of elements
            if elems <= 0:
                elems = int((len(rawdata) - offset) / basictypes[ftype]["size"])

            # build format string
            fmt = endian + str(elems) + basictypes[ftype]["fmt"]

            # parse data
            value = struct.unpack_from(fmt, rawdata, offset)

            # set field
            if ftype == "STRING":
                rec[fieldname] = ctypes.create_string_buffer(value[0]).value.decode('ascii')
            elif is_array:
                rec[fieldname] = value
            else:
                rec[fieldname] = value[0]

        # decode user type
        else:

            # populate record definition (if needed) #
            subrecdef = __populate(ftype)

            # check if array
            is_array = not (elems == 1)

            # get number of elements
            if elems <= 0:
                elems = int((len(rawdata) - offset) / subrecdef["__datasize"])

            # return parsed data
            if is_array:
                rec[fieldname] = []
                for e in range(elems):
                    rec[fieldname].append(__decode_native(ftype, rawdata[offset:]))
                    offset += subrecdef["__datasize"]
            else:
                rec[fieldname] = __decode_native(ftype, rawdata[offset:])

    # return record #
    return rec

#
#  __parse_native
#
def __parse_native(data, callbacks):
    """
    data: request response
    """
    recs = []

    rec_hdr_size = 8
    rec_size_index = 0
    rec_size_rsps = []

    rec_size = 0
    rec_index = 0
    rec_rsps = []

    duration = 0.0

    for line in data:

        # Capture Start Time (for duration)
        tstart = time.perf_counter()

        # Process Line Read
        i = 0
        while i < len(line):

            # Parse Record Size
            if(rec_size_index < rec_hdr_size):
                bytes_available = len(line) - i
                bytes_remaining = rec_hdr_size - rec_size_index
                bytes_to_append = min(bytes_available, bytes_remaining)
                rec_size_rsps.append(line[i:i+bytes_to_append])
                rec_size_index += bytes_to_append
                if(rec_size_index >= rec_hdr_size):
                    raw = b''.join(rec_size_rsps)
                    rec_version, rec_type_size, rec_data_size = struct.unpack('>hhi', raw)
                    if rec_version != 2:
                        raise FatalError("Invalid record format: %d" % (rec_version))
                    rec_size = rec_type_size + rec_data_size
                    rec_size_rsps.clear()
                i += bytes_to_append

            # Parse Record
            elif(rec_size > 0):
                bytes_available = len(line) - i
                bytes_remaining = rec_size - rec_index
                bytes_to_append = min(bytes_available, bytes_remaining)
                rec_rsps.append(line[i:i+bytes_to_append])
                rec_index += bytes_to_append
                if(rec_index >= rec_size):
                    # Decode Record
                    rawbits = b''.join(rec_rsps)
                    rectype = ctypes.create_string_buffer(rawbits).value.decode('ascii')
                    rawdata = rawbits[len(rectype) + 1:]
                    rec     = __decode_native(rectype, rawdata)
                    if rectype == "conrec":
                        # parse records contained in record
                        buffer = __BufferSource(rawdata[rec["start"]:])
                        contained_recs = __parse_native(buffer, callbacks)
                        recs += contained_recs
                    else:
                        if callbacks != None and rectype in callbacks:
                            # Execute Call-Back on Record
                            callbacks[rectype](rec)
                        else:
                            # Append Record
                            recs.append(rec)
                    # Reset Record Parsing
                    rec_rsps.clear()
                    rec_size_index = 0
                    rec_size = 0
                    rec_index = 0
                i += bytes_to_append

            # Zero Sized Record
            else:
                rec_size_index = 0
                rec_index = 0

        # Capture Duration
        duration = duration + (time.perf_counter() - tstart)

    # Update Timing Profile
    profiles[__parse_native.__name__] = duration

    return recs


###############################################################################
# Overriding DNS
###############################################################################

#
# local_dns - global dictionary of DNS entries to override
#
local_dns = {}

#
# __initdns - resets the global dictionary
#
def __initdns():
    local_dns.clear()

#
# __dnsoverridden - checks if url is already overridden
#
def __dnsoverridden():
    global service_org, service_url
    url = service_org + "." + service_url
    return url.lower() in local_dns

#
# __jamdns - override the dns entry
#
def __jamdns():
    global service_url, service_org, request_timeout
    url = service_org + "." + service_url
    headers = buildauthheader()
    host = "https://ps." + service_url + "/api/org_ip_adr/" + service_org + "/"
    rsps = session.get(host, headers=headers, timeout=request_timeout).json()
    if rsps["status"] == "SUCCESS":
        ipaddr = rsps["ip_address"]
        local_dns[url.lower()] = ipaddr
        logger.info("Overriding DNS for {} with {}".format(url, ipaddr))

#
# __override_getaddrinfo - replaces the socket library callback
#
socket_getaddrinfo = socket.getaddrinfo
def __override_getaddrinfo(*args):
    url = args[0].lower()
    if url in local_dns:
        logger.info("getaddrinfo returned {} for {}".format(local_dns[url], url))
        return socket_getaddrinfo(local_dns[url], *args[1:])
    else:
        return socket_getaddrinfo(*args)
socket.getaddrinfo = __override_getaddrinfo


###############################################################################
# Default Record Processing
###############################################################################

#
#  __logeventrec
#
def __logeventrec(rec):
    eventlogger[rec['level']]('Log <%s, %d>: %s' % (rec["name"], rec["time"], rec["attr"]))

#
#  __exceptrec
#
def __exceptrec(rec):
    if rec["code"] == exceptioncodes["SIMPLIFY"]:
        raise RetryRequest("cmr simplification requested")
    elif rec["code"] < 0:
        eventlogger[rec["level"]]("Exception <%d>: %s", rec["code"], rec["text"])
    else:
        eventlogger[rec["level"]]("%s", rec["text"])

#
#  _arrowrec
#
def __arrowrec(rec):
    global arrow_file_table
    try :
        filename = rec["filename"]
        if rec["__rectype"] == 'arrowrec.meta':
            if filename in arrow_file_table:
                raise FatalError(f'File transfer already in progress for {filename}')
            logger.info(f'Writing arrow file: {filename}')
            arrow_file_table[filename] = { "fp": open(filename, "wb"), "size": rec["size"], "progress": 0 }
        elif rec["__rectype"] == 'arrowrec.eof':
            logger.info(f'Checksum of arrow file: {rec["checksum"]}')
        else: # rec["__rectype"] == 'arrowrec.data'
            data = rec['data']
            file = arrow_file_table[filename]
            file["fp"].write(bytearray(data))
            file["progress"] += len(data)
            if file["progress"] >= file["size"]:
                file["fp"].close()
                logger.info(f'Closing arrow file: {filename}')
                del arrow_file_table[filename]
    except Exception as e:
        raise FatalError(f'Failed to process arrow file: {e}')

#
#  Globals
#
__callbacks = {'eventrec': __logeventrec, 'exceptrec': __exceptrec, 'arrowrec.meta': __arrowrec, 'arrowrec.data': __arrowrec, 'arrowrec.eof': __arrowrec}


###############################################################################
# INTERNAL APIs
###############################################################################

#
#  buildauthheader
#
def buildauthheader(force_refresh=False):
    """
    Build authentication header for use with provisioning system
    """
    global service_url, ps_access_token, ps_refresh_token, ps_token_exp
    headers = None
    if ps_access_token:
        # Check if Refresh Needed
        if time.time() > ps_token_exp or force_refresh:
            host = "https://ps." + service_url + "/api/org_token/refresh/"
            rqst = {"refresh": ps_refresh_token}
            hdrs = {'Content-Type': 'application/json', 'Authorization': 'Bearer ' + ps_access_token}
            rsps = session.post(host, data=json.dumps(rqst), headers=hdrs, timeout=request_timeout).json()
            ps_refresh_token = rsps["refresh"]
            ps_access_token = rsps["access"]
            ps_token_exp =  time.time() + (float(rsps["access_lifetime"]) / 2)
        # Build Authentication Header
        headers = {'Authorization': 'Bearer ' + ps_access_token}
    return headers

#
# GeoDataFrame to Polygon
#
def gdf2poly(gdf):

    # latch start time
    tstart = time.perf_counter()

    # pull out coordinates
    hull = gdf.unary_union.convex_hull
    polygon = [{"lon": coord[0], "lat": coord[1]} for coord in list(hull.exterior.coords)]

    # determine winding of polygon #
    #              (x2               -    x1)             *    (y2               +    y1)
    wind = sum([(polygon[i+1]["lon"] - polygon[i]["lon"]) * (polygon[i+1]["lat"] + polygon[i]["lat"]) for i in range(len(polygon) - 1)])
    if wind > 0:
        # reverse direction (make counter-clockwise) #
        ccw_poly = []
        for i in range(len(polygon), 0, -1):
            ccw_poly.append(polygon[i - 1])
        # replace region with counter-clockwise version #
        polygon = ccw_poly

    # Update Profile
    profiles[gdf2poly.__name__] = time.perf_counter() - tstart

    # return polygon
    return polygon

#
#  Create Empty GeoDataFrame
#
def emptyframe(**kwargs):
    # set default keyword arguments
    kwargs['crs'] = SLIDERULE_EPSG
    return geopandas.GeoDataFrame(geometry=geopandas.points_from_xy([], [], []), crs=kwargs['crs'])

#
# Process Output File
#
def procoutputfile(parm, rsps):
    output = parm["output"]
    if "path" in output:
        path = output["path"]
    else:
        path = None
    # Check If Remote Record Is In Responses
    for rsp in rsps:
        if 'arrowrec.remote' == rsp['__rectype']:
            path = rsp['url']
            break
    # Handle Local Files
    if "open_on_complete" in output and output["open_on_complete"]:
        if output["format"] == "parquet":
            if "as_geo" in output and output["as_geo"]:
                # Return GeoParquet File as GeoDataFrame
                return geopandas.read_parquet(path)
            else:
                # Return Parquet File as DataFrame
                return geopandas.pd.read_parquet(path)
        elif output["format"] == "geoparquet":
            # Return Parquet File as DataFrame
            return geopandas.pd.read_parquet(path)
        elif output["format"] == "feather":
            # Return Feather File as DataFrame
            return geopandas.pd.read_feather(path)
        elif output["format"] == "csv":
            # Return CSV File as DataFrame
            return geopandas.pd.read_csv(path)
        else:
            raise FatalError('unsupported output format: %s' % (output["format"]))
    else:
        # Return Parquet Filename
        return path

#
#  Get Values from Raw Buffer
#
def getvalues(data, dtype, size, num_elements=0):
    """
    data:   tuple of bytes
    dtype:  element of codedtype
    size:   bytes in data
    """

    raw = bytes(data)
    datatype = basictypes[codedtype2str[dtype]]["nptype"]
    if num_elements == 0: # dynamically determine number of elements
        num_elements = int(size / numpy.dtype(datatype).itemsize)
    slicesize = num_elements * numpy.dtype(datatype).itemsize # truncates partial bytes
    values = numpy.frombuffer(raw[:slicesize], dtype=datatype, count=num_elements)

    return values

#
#  Dictionary to GeoDataFrame
#
def todataframe(columns, time_key="time", lon_key="longitude", lat_key="latitude", height_key=None, **kwargs):

    # Latch Start Time
    tstart = time.perf_counter()

    # Set Default Keyword Arguments
    kwargs['index_key'] = "time"
    kwargs['crs'] = SLIDERULE_EPSG

    # Check Empty Columns
    if len(columns) <= 0:
        return emptyframe(**kwargs)

    # Generate Time Column
    columns['time'] = columns[time_key].astype('datetime64[ns]')

    # Generate Geometry Column
    # 3D point geometry
    # This enables 3D CRS transformations using the to_crs() method
    if height_key == None:
        geometry = geopandas.points_from_xy(columns[lon_key], columns[lat_key])
    else:
        geometry = geopandas.points_from_xy(columns[lon_key], columns[lat_key], columns[height_key])
    del columns[lon_key]
    del columns[lat_key]

    # Create Pandas DataFrame object
    if type(columns) == dict:
        df = geopandas.pd.DataFrame(columns)
    else:
        df = columns

    # Build GeoDataFrame (default geometry is crs=SLIDERULE_EPSG)
    gdf = geopandas.GeoDataFrame(df, geometry=geometry, crs=kwargs['crs'])

    # Set index (default is Timestamp), can add `verify_integrity=True` to check for duplicates
    # Can do this during DataFrame creation, but this allows input argument for desired column
    gdf.set_index(kwargs['index_key'], inplace=True)

    # Sort values for reproducible output despite async processing
    gdf.sort_index(inplace=True)

    # Update Profile
    profiles[todataframe.__name__] = time.perf_counter() - tstart

    # Return GeoDataFrame
    return gdf

#
# Simplify Polygon
#
def simplifypolygon(parm):
    if "parms" not in parm:
        return

    if "cmr" in parm["parms"]:
        polygon = parm["parms"]["cmr"]["polygon"]
    elif "poly" in parm["parms"]:
        polygon = parm["parms"]["poly"]
    else:
        return

    tolerance = 0.01
    raw_multi_polygon = [[(tuple([(c['lon'], c['lat']) for c in polygon]), [])]]
    shape = MultiPolygon(*raw_multi_polygon)
    buffered_shape = shape.buffer(tolerance)
    simplified_shape = buffered_shape.simplify(tolerance)
    simplified_coords = list(simplified_shape.exterior.coords)

    simplified_polygon = []
    for coord in simplified_coords:
        point = {"lon": coord[0], "lat": coord[1]}
        simplified_polygon.insert(0, point)

    if "cmr" not in parm["parms"]:
        parm["parms"]["cmr"] = {}
    parm["parms"]["cmr"]["polygon"] = simplified_polygon

    logger.warning('Using simplified polygon (for CMR request only!), {} points using tolerance of {}'.format(len(simplified_coords), tolerance))

#
# rethrow
#
def rethrow():
    global rethrow_exceptions
    return rethrow_exceptions

###############################################################################
# APIs
###############################################################################

#
#  Initialize
#
def init (
    url=PUBLIC_URL,
    verbose=False,
    loglevel=logging.INFO,
    organization=0,
    desired_nodes=None,
    time_to_live=60,
    bypass_dns=False,
    plugins=[],
    trust_env=DEFAULT_TRUST_ENV,
    log_handler=None,
    rethrow=None
):
    '''
    Initializes the Python client for use with SlideRule, and should be called before other ICESat-2 API calls.
    This function is a wrapper for a handful of sliderule functions that would otherwise all have to be called in order to initialize the client.

    Parameters
    ----------
        url:            str
                        the IP address or hostname of the SlideRule service (slidereearth.io by default)
        verbose:        bool
                        sets up console logger as a convenience to user so all logs are printed to screen
        loglevel:       int
                        minimum severity of log message to output
        organization:   str
                        SlideRule provisioning system organization the user belongs to (see sliderule.authenticate for details)
        desired_nodes:  int
                        requested number of processing nodes in the cluster
        time_to_live:   int
                        minimum number of minutes the desired number of nodes should be present in the cluster
        bypass_dns:     bool
                        if true then the ip address for the cluster is retrieved from the provisioning system and used directly
        plugins:        list
                        names of the plugins that need to be available on the server
        log_handler:    logger
                        user provided logging handler
        rethrow:        bool
                        configure client to throw any exceptions it encounters instead of automatic retries

    Returns
    -------
    bool
        Status of version check

    Examples
    --------
        >>> import sliderule
        >>> sliderule.init()
    '''
    global session, logger, rethrow_exceptions
    # massage function parameters
    if organization == 0:
        organization = PUBLIC_ORG
    # reconfigure session (if necessary)
    if session.trust_env != trust_env:
        session.trust_env = trust_env
    # configure logging
    if log_handler != None:
        logger.addHandler(log_handler)
    set_verbose(verbose, loglevel)
    # configure client
    if rethrow != None:
        rethrow_exceptions = rethrow
    set_url(url) # configure domain
    authenticate(organization) # configure credentials (if any) for organization
    scaleout(desired_nodes, time_to_live, bypass_dns) # set cluster to desired number of nodes (if permitted based on credentials)
    return check_version(plugins=plugins) # verify compatibility between client and server versions

#
#  source
#
def source (api, parm={}, stream=False, callbacks={}, path="/source", silence=False):
    '''
    Perform API call to SlideRule service

    Parameters
    ----------
        api:        str
                    name of the SlideRule endpoint
        parm:       dict
                    dictionary of request parameters
        stream:     bool
                    whether the request is a **normal** service or a **stream** service (see `De-serialization </web/rtd/user_guide/SlideRule.html#de-serialization>`_ for more details)
        callbacks:  dict
                    record type callbacks (advanced use)
        path:       str
                    path to api being requested
        silence:    bool
                    whether error log messages should be generated

    Returns
    -------
    dictionary
        response data

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("slideruleearth.io")
        >>> rqst = {
        ...     "time": "NOW",
        ...     "input": "NOW",
        ...     "output": "GPS"
        ... }
        >>> rsps = sliderule.source("time", rqst)
        >>> print(rsps)
        {'time': 1300556199523.0, 'format': 'GPS'}
    '''
    global service_url, service_org, rethrow_exceptions
    rsps = {}
    headers = None
    # Build Callbacks
    for c in __callbacks:
        if c not in callbacks:
            callbacks[c] = __callbacks[c]
    # Attempt Request
    complete = False
    attempts = 3
    while not complete and attempts > 0:
        attempts -= 1
        try:
            # Dump Parameters to Request JSON String
            rqst = json.dumps(parm)
            # Construct Request URL and Authorization
            if service_org:
                url = 'https://%s.%s%s/%s' % (service_org, service_url, path, api)
                headers = buildauthheader()
            else:
                url = 'http://%s%s/%s' % (service_url, path, api)
            # Perform Request
            if not stream:
                data = session.get(url, data=rqst, headers=headers, timeout=request_timeout)
            else:
                data = session.post(url, data=rqst, headers=headers, timeout=request_timeout, stream=True)
            data.raise_for_status()
            # Parse Response
            stream = __StreamSource(data)
            format = data.headers['Content-Type']
            if format == 'text/plain':
                rsps = __parse_json(stream)
            elif format == 'application/json':
                rsps = __parse_json(stream)
            elif format == 'application/octet-stream':
                rsps = __parse_native(stream, callbacks)
            else:
                raise FatalError('unsupported content type: %s' % (format))
            # Success
            complete = True
        except RetryRequest as e:
            logger.info("Retry requested by {}: {}".format(url, e))
            simplifypolygon(parm)
        except requests.exceptions.SSLError as e:
            logger.debug("Exception in request to {}: {}".format(url, e))
            if rethrow_exceptions:
                raise
            if not silence:
                logger.error("Unable to verify SSL certificate for {} ...retrying request".format(url))
        except requests.ConnectionError as e:
            logger.debug("Exception in request to {}: {}".format(url, e))
            if rethrow_exceptions:
                raise
            if not silence:
                logger.error("Connection error to endpoint {} ...retrying request".format(url))
        except requests.Timeout as e:
            logger.debug("Exception in request to {}: {}".format(url, e))
            if rethrow_exceptions:
                raise
            if not silence:
                logger.error("Timed-out waiting for response from endpoint {} ...retrying request".format(url))
        except requests.exceptions.ChunkedEncodingError as e:
            logger.debug("Exception in request to {}: {}".format(url, e))
            if rethrow_exceptions:
                raise
            if not silence:
                logger.error("Unexpected termination of response from endpoint {} ...retrying request".format(url))
        except requests.HTTPError as e:
            logger.debug("Exception in request to {}: {}".format(url, e))
            if e.response.status_code == 503:
                raise TransientError("Server experiencing heavy load, stalling on request to {}".format(url))
            else:
                raise FatalError("HTTP error {} from endpoint {}".format(e.response.status_code, url))
        except:
            raise
    # Check Success
    if not complete:
        raise FatalError("Unable to complete request due to errors")
    # Return Response
    return rsps

#
#  set_url
#
def set_url (url):
    '''
    Configure sliderule package with URL of service

    Parameters
    ----------
        urls:   str
                IP address or hostname of SlideRule service (note, there is a special case where the url is provided as a list of strings
                instead of just a string; when a list is provided, the client hardcodes the set of servers that are used to process requests
                to the exact set provided; this is used for testing and for local installations and can be ignored by most users)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("service.my-sliderule-server.org")
    '''
    global service_url
    service_url = url
    logger.info(f'Setting URL to {service_url}')

#
#  set_verbose
#
def set_verbose (enable, loglevel=logging.INFO):
    '''
    Sets up a console logger to print log messages to screen

    If you want more control over the behavior of the log messages being captured, do not call this function but instead
    create and configure a Python log handler of your own and attach it to `sliderule.logger`.

    Parameters
    ----------
        enable:     bool
                    True: creates console logger if it doesn't exist, False: destroys console logger if it does exist

        loglevel:       int
                        minimum severity of log message to output

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_verbose(True, loglevel=logging.INFO)
    '''
    global console, logger
    # massage loglevel parameter if passed in as a string
    if loglevel == "DEBUG":
        loglevel = logging.DEBUG
    elif loglevel == "INFO":
        loglevel = logging.INFO
    elif loglevel == "WARNING":
        loglevel = logging.WARNING
    elif loglevel == "WARN":
        loglevel = logging.WARN
    elif loglevel == "ERROR":
        loglevel = logging.ERROR
    elif loglevel == "FATAL":
        loglevel = logging.FATAL
    elif loglevel == "CRITICAL":
        loglevel = logging.CRITICAL
    # enable/disable logging to console
    if (enable == True) and (console == None):
        console = logging.StreamHandler()
        logger.addHandler(console)
    elif (enable == False) and (console != None):
        logger.removeHandler(console)
        console = None
    # always set level to requested
    logger.setLevel(loglevel)
    if console != None:
        console.setLevel(loglevel)

#
# set_rqst_timeout
#
def set_rqst_timeout (timeout):
    '''
    Sets the TCP/IP connection and reading timeouts for future requests made to sliderule servers.
    Setting it lower means the client will failover more quickly, but may generate false positives if a processing request stalls or takes a long time returning data.
    Setting it higher means the client will wait longer before designating it a failed request which in the presence of a persistent failure means it will take longer for the client to remove the node from its available servers list.

    Parameters
    ----------
        timeout:    tuple
                    (<connection timeout in seconds>, <read timeout in seconds>)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_rqst_timeout((10, 60))
    '''
    global request_timeout
    if type(timeout) == tuple:
        request_timeout = timeout
    else:
        raise FatalError('timeout must be a tuple (<connection timeout>, <read timeout>)')

#
# set_processing_flags
#
def set_processing_flags (aux=True):
    '''
    Sets flags used when processing the record definitions

    Parameters
    ----------
        aux:    bool
                decode auxiliary fields

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_processing_flags(aux=False)
    '''
    global decode_aux
    if type(aux) == bool:
        decode_aux = aux
    else:
        raise FatalError('aux must be a boolean')


#
# update_available_servers
#
def update_available_servers (desired_nodes=None, time_to_live=None):
    '''
    Manages the number of servers in the cluster.
    If the desired_nodes parameter is set, then a request is made to change the number of servers in the cluster to the number specified.
    In all cases, the number of nodes currently running in the cluster are returned - even if desired_nodes is set;
    subsequent calls to this function is needed to check when the current number of nodes matches the desired_nodes.

    Parameters
    ----------
        desired_nodes:  int
                        the desired number of nodes in the cluster
        time_to_live:   int
                        number of minutes for the desired nodes to run

    Returns
    -------
    int
        number of nodes currently in the cluster
    int
        number of nodes available for work in the cluster

    Examples
    --------
        >>> import sliderule
        >>> num_servers, max_workers = sliderule.update_available_servers(10)
    '''
    global service_url, service_org, request_timeout, rethrow_exceptions
    requested_nodes = 0

    # Update number of nodes
    if type(desired_nodes) == int:
        rsps_body = {}
        requested_nodes = desired_nodes
        headers = buildauthheader()

        # Get boundaries of cluster and calculate nodes to request
        try:
            host = "https://ps." + service_url + "/api/org_num_nodes/" + service_org + "/"
            rsps = session.get(host, headers=headers, timeout=request_timeout)
            rsps_body = rsps.json()
            rsps.raise_for_status()
            requested_nodes = max(min(desired_nodes, rsps_body["max_nodes"]), rsps_body["min_nodes"])
            if requested_nodes != desired_nodes:
                logger.info("Provisioning system desired nodes overridden to {}".format(requested_nodes))
        except requests.exceptions.HTTPError as e:
            logger.info('{}'.format(e))
            logger.info('Provisioning system status request returned error => {}'.format(rsps_body["error_msg"]))
            if rethrow_exceptions:
                raise

        # Request number of nodes in cluster
        try:
            if type(time_to_live) == int:
                host = "https://ps." + service_url + "/api/desired_org_num_nodes_ttl/" + service_org + "/" + str(requested_nodes) + "/" + str(time_to_live) + "/"
                rsps = session.post(host, headers=headers, timeout=request_timeout)
            else:
                host = "https://ps." + service_url + "/api/desired_org_num_nodes/" + service_org + "/" + str(requested_nodes) + "/"
                rsps = session.put(host, headers=headers, timeout=request_timeout)
            rsps_body = rsps.json()
            rsps.raise_for_status()
        except requests.exceptions.HTTPError as e:
            logger.info('{}'.format(e))
            logger.error('Provisioning system update request error => {}'.format(rsps_body["error_msg"]))
            if rethrow_exceptions:
                raise

    # Get number of nodes currently registered
    try:
        rsps = source("status", parm={"service":"sliderule"}, path="/discovery", silence=True)
        available_servers = rsps["nodes"]
    except FatalError as e:
        logger.debug("Failed to retrieve number of nodes registered: {}".format(e))
        available_servers = 0

    return available_servers, requested_nodes

#
# scaleout
#
def scaleout(desired_nodes, time_to_live, bypass_dns):
    '''
    Scale the cluster and wait for cluster to reach desired state

    Parameters
    ----------
        desired_nodes:  int
                        the desired number of nodes in the cluster
        time_to_live:   int
                        number of minutes for the desired nodes to run
        bypass_dns:     bool
                        the cluster ip address is retrieved from the provisioning system and used directly

    Examples
    --------
        >>> import sliderule
        >>> sliderule.scaleout(4, 300, False)
    '''
    if desired_nodes is None:
        return # nothing needs to be done
    if desired_nodes < 0:
        raise FatalError("Number of desired nodes must be greater than zero ({})".format(desired_nodes))
    # Initialize DNS
    __initdns() # clear cache of DNS lookups for clusters
    if bypass_dns:
        __jamdns() # use ip address for cluster
    # Send Initial Request for Desired Cluster State
    start = time.time()
    available_nodes,requested_nodes = update_available_servers(desired_nodes=desired_nodes, time_to_live=time_to_live)
    scale_up_needed = False
    # Wait for Cluster to Reach Desired State
    while available_nodes < requested_nodes:
        scale_up_needed = True
        logger.info("Waiting while cluster scales to desired capacity (currently at {} nodes, desired is {} nodes)... {} seconds".format(available_nodes, desired_nodes, int(time.time() - start)))
        time.sleep(10.0)
        available_nodes,_ = update_available_servers()
        # Override DNS if Cluster is Starting
        if available_nodes == 0 and not __dnsoverridden():
            __jamdns()
        # Timeout Occurred
        if int(time.time() - start) > MAX_PS_CLUSTER_WAIT_SECS:
            logger.error("Maximum time allowed waiting for cluster has been exceeded")
            break
    # Log Final Message if Cluster Needed State Change
    if scale_up_needed:
        logger.info("Cluster has reached capacity of {} nodes... {} seconds".format(available_nodes, int(time.time() - start)))

#
# authenticate
#
def authenticate (ps_organization, ps_username=None, ps_password=None, github_token=None):
    '''
    Authenticate to SlideRule Provisioning System
    The username and password can be provided the following way in order of priority:
    (1) The passed in arguments `github_token` or `ps_username` and `ps_password`;
    (2) The O.S. environment variables `PS_GITHUB_TOKEN` or `PS_USERNAME` and `PS_PASSWORD`;
    (3) The `ps.<url>` entry in the .netrc file in your home directory

    Parameters
    ----------
        ps_organization:    str
                            name of the SlideRule organization the user belongs to

        ps_username:        str
                            SlideRule provisioning system account name

        ps_password:        str
                            SlideRule provisioning system account password

        github_token:       str
                            GitHub access token (minimum scope/permissions require)

    Returns
    -------
    status
        True of successful, False if unsuccessful

    Examples
    --------
        >>> import sliderule
        >>> sliderule.authenticate("myorg")
        True
    '''
    global service_org, ps_refresh_token, ps_access_token, ps_token_exp
    login_status = False
    ps_url = "ps." + service_url

    # set organization on any authentication request
    service_org = ps_organization

    # check for direct or public access
    if service_org == None:
        return True

    # attempt retrieving from environment
    if not github_token and not ps_username and not ps_password:
        github_token = os.environ.get("PS_GITHUB_TOKEN")
        ps_username = os.environ.get("PS_USERNAME")
        ps_password = os.environ.get("PS_PASSWORD")

    # attempt retrieving from netrc file
    if not github_token and not ps_username and not ps_password:
        try:
            netrc_file = netrc.netrc()
            login_credentials = netrc_file.hosts[ps_url]
            ps_username = login_credentials[0]
            ps_password = login_credentials[2]
        except Exception as e:
            if ps_organization != PUBLIC_ORG:
                logger.warning("Unable to retrieve username and password from netrc file for machine: {}".format(e))

    # build authentication request
    user = None
    if github_token:
        user = "github"
        rqst = {"org_name": ps_organization, "access_token": github_token}
        headers = {'Content-Type': 'application/json'}
        api = "https://" + ps_url + "/api/org_token_github/"
    elif ps_username or ps_password:
        user = "local"
        rqst = {"username": ps_username, "password": ps_password, "org_name": ps_organization}
        headers = {'Content-Type': 'application/json'}
        api = "https://" + ps_url + "/api/org_token/"

    # authenticate to provisioning system
    if user:
        try:
            rsps = session.post(api, data=json.dumps(rqst), headers=headers, timeout=request_timeout)
            rsps.raise_for_status()
            rsps = rsps.json()
            ps_refresh_token = rsps["refresh"]
            ps_access_token = rsps["access"]
            ps_token_exp =  time.time() + (float(rsps["access_lifetime"]) / 2)
            login_status = True
        except:
            if ps_organization != PUBLIC_ORG:
                logger.error("Unable to authenticate %s user to %s" % (user, api))

    # return login status
    if ps_organization != PUBLIC_ORG:
        logger.info(f'Login status to {service_url}/{service_org}: {login_status and "success" or "failure"}')
    return login_status

#
# gps2utc
#
def gps2utc (gps_time, as_str=True):
    '''
    Convert a GPS based time returned from SlideRule into a UTC time.

    Parameters
    ----------
        gps_time:   float
                    number of seconds since GPS epoch (January 6, 1980)
        as_str:     bool
                    if True, returns the time as a string; if False, returns the time as datatime object

    Returns
    -------
    datetime
        UTC time (i.e. GMT, or Zulu time)

    Examples
    --------
        >>> import sliderule
        >>> sliderule.gps2utc(1235331234)
        '2019-02-27 19:34:03'
    '''
    rsps = source("time", {"time": int(gps_time * 1000), "input": "GPS", "output": "DATE"})
    if as_str:
        return rsps["time"]
    else:
        return datetime.strptime(rsps["time"], '%Y-%m-%dT%H:%M:%SZ')
#
# get_definition
#
def get_definition (rectype, fieldname):
    '''
    Get the underlying format specification of a field in a return record.

    Parameters
    ----------
    rectype:    str
                the name of the type of the record (i.e. "atl03rec")
    fieldname:  str
                the name of the record field (i.e. "cycle")

    Returns
    -------
    dict
        description of each field; see the `sliderule.basictypes` variable for different field types

    Examples
    --------
        >>> import sliderule
        >>> sliderule.set_url("slideruleearth.io")
        >>> sliderule.get_definition("atl03rec", "cycle")
        {'fmt': 'H', 'size': 2, 'nptype': <class 'numpy.uint16'>}
    '''
    recdef = __populate(rectype)
    if fieldname in recdef and recdef[fieldname]["type"] in basictypes:
        return basictypes[recdef[fieldname]["type"]]
    else:
        return {}

#
# get_version
#
def get_version ():
    '''
    Get the version information for the running servers and Python client

    Returns
    -------
    dict
        dictionary of version information
    '''
    global service_org
    rsps = source("version", {})
    rsps["client"] = {"version": version.full_version}
    rsps["organization"] = service_org
    return rsps

#
# check_version
#
def check_version (plugins=[]):
    '''
    Check that the version of the client matches the version of the server and any additionally requested plugins

    Parameters
    ----------
    plugins:    list
                list of package names (as strings) to check the version on

    Returns
    -------
    bool
        True if at least minor version matches; False if major or minor version doesn't match
    '''
    status = True
    info = get_version()
    # populate version info
    versions = {}
    for entity in ['server', 'client'] + plugins:
        s = info[entity]['version'][1:].split('.')
        versions[entity] = (int(s[0]), int(s[1]), int(s[2]))
    # check major version mismatches
    if versions['server'][0] != versions['client'][0]:
        raise RuntimeError("Client (version {}) is incompatible with the server (version {})".format(versions['client'], versions['server']))
    else:
        for pkg in plugins:
            if versions[pkg][0] != versions['client'][0]:
                raise RuntimeError("Client (version {}) is incompatible with the {} plugin (version {})".format(versions['client'], pkg, versions[pkg]))
    # check minor version mismatches
    if versions['server'][1] > versions['client'][1]:
        logger.warning("Client (version {}) is out of date with the server (version {})".format(versions['client'], versions['server']))
        status = False
    else:
        for pkg in plugins:
            if versions[pkg][1] > versions['client'][1]:
                logger.warning("Client (version {}) is out of date with the {} plugin (version {})".format(versions['client'], pkg, versions['server']))
                status = False
    # return if version check is successful
    return status

#
# Format Region Specification
#
def toregion(source, tolerance=0.0, cellsize=0.01, n_clusters=1):
    '''
    Convert a GeoJSON/Shapefile/GeoDataFrame/list representation of a set of geospatial regions into a list of lat,lon coordinates and raster image recognized by SlideRule

    Parameters
    ----------
        source:     str
                    file name of GeoJSON formatted regions of interest, file **must** have name with the .geojson suffix
                    file name of ESRI Shapefile formatted regions of interest, file **must** have name with the .shp suffix
                    GeoDataFrame of region of interest
                    list of longitude,latitude pairs forming a polygon (e.g. [lat1, lon1, lat2, lon2, lat3, lon3, lat1, lon1])
                    list of longitude,latitude pairs forming a bounding box (e.g. [lat1, lon1, lat2, lon2])
        tolerance:  float
                    tolerance used to simplify complex shapes so that the number of points is less than the limit (a tolerance of 0.001 typically works for most complex shapes)
        cellsize:   float
                    size of pixel in degrees used to create the raster image of the polygon
        n_clusters: int
                    number of clusters of polygons to create when breaking up the request to CMR

    Returns
    -------
    dict
        a list of longitudes and latitudes containing the region of interest that can be used for the **poly** and **raster** parameters in a processing request to SlideRule.

        region = {

            "gdf": <GeoDataFrame of region>

            "poly": [{"lat": <lat1>, "lon": <lon1> }, ...],

            "raster": {"data": <geojson file as string>,

            "clusters": [[{"lat": <lat1>, "lon": <lon1>}, ...], [{"lat": <lat1>, "lon": <lon1>}, ...]] }

    Examples
    --------
    >>> import sliderule, json
    >>> region = sliderule.toregion("tests/data/grandmesa.geojson")
    >>> print(json.dumps(region["poly"], indent=4))
    [
        {
            "lon": -108.20772968780051,
            "lat": 38.8232055291981
        },
        {
            "lon": -108.07460164311031,
            "lat": 38.8475137825863
        },
        {
            "lon": -107.72839858755752,
            "lat": 39.01510930230633
        },
        {
            "lon": -107.78724142490994,
            "lat": 39.195630349659986
        },
        {
            "lon": -108.17287000970857,
            "lat": 39.15920066396116
        },
        {
            "lon": -108.31168256553767,
            "lat": 39.13757646212944
        },
        {
            "lon": -108.34115668325224,
            "lat": 39.03758987613325
        },
        {
            "lon": -108.2878686387796,
            "lat": 38.89051431295789
        },
        {
            "lon": -108.20772968780051,
            "lat": 38.8232055291981
        }
    ]
    '''

    tstart = time.perf_counter()
    tempfile = "temp.geojson"

    if isinstance(source, geopandas.GeoDataFrame):
        # user provided GeoDataFrame instead of a file
        gdf = source
        # Convert to geojson file
        gdf.to_file(tempfile, driver="GeoJSON")
        with open(tempfile, mode='rt') as file:
            datafile = file.read()
        os.remove(tempfile)

    elif isinstance(source, Polygon):
        gdf = geopandas.GeoDataFrame(geometry=[source], crs=EPSG_WGS84)
        gdf.to_file(tempfile, driver="GeoJSON")
        with open(tempfile, mode='rt') as file:
            datafile = file.read()
        os.remove(tempfile)

    elif isinstance(source, list) and (len(source) >= 4) and (len(source) % 2 == 0):
        # create lat/lon lists
        if len(source) == 4: # bounding box
            lons = [source[0], source[2], source[2], source[0], source[0]]
            lats = [source[1], source[1], source[3], source[3], source[1]]
        elif len(source) > 4: # polygon list
            lons = [source[i] for i in range(1,len(source),2)]
            lats = [source[i] for i in range(0,len(source),2)]

        # create geodataframe
        p = Polygon([point for point in zip(lons, lats)])
        gdf = geopandas.GeoDataFrame(geometry=[p], crs=EPSG_WGS84)

        # Convert to geojson file
        gdf.to_file(tempfile, driver="GeoJSON")
        with open(tempfile, mode='rt') as file:
            datafile = file.read()
        os.remove(tempfile)

    elif isinstance(source, str) and (source.find(".shp") > 1):
        # create geodataframe
        gdf = geopandas.read_file(source)
        # Convert to geojson file
        gdf.to_file(tempfile, driver="GeoJSON")
        with open(tempfile, mode='rt') as file:
            datafile = file.read()
        os.remove(tempfile)

    elif isinstance(source, str) and (source.find(".geojson") > 1):
        # create geodataframe
        gdf = geopandas.read_file(source)
        with open(source, mode='rt') as file:
            datafile = file.read()

    else:
        raise FatalError("incorrect filetype: please use a .geojson, .shp, or a geodataframe")


    # If user provided raster we don't have gdf, geopandas cannot easily convert it
    polygon = clusters = None
    if gdf is not None:
        # simplify polygon
        if(tolerance > 0.0):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                gdf = gdf.buffer(tolerance)
                gdf = gdf.simplify(tolerance)

        # generate polygon
        polygon = gdf2poly(gdf)

        # generate clusters
        clusters = []
        if n_clusters > 1:
            if clustering_enabled:
                # pull out centroids of each geometry object
                if "CenLon" in gdf and "CenLat" in gdf:
                    X = numpy.column_stack((gdf["CenLon"], gdf["CenLat"]))
                else:
                    s = gdf.centroid
                    X = numpy.column_stack((s.x, s.y))
                # run k means clustering algorithm against polygons in gdf
                kmeans = KMeans(n_clusters=n_clusters, init='k-means++', random_state=5, max_iter=400)
                y_kmeans = kmeans.fit_predict(X)
                k = geopandas.pd.DataFrame(y_kmeans, columns=['cluster'])
                gdf = gdf.join(k)
                # build polygon for each cluster
                for n in range(n_clusters):
                    c_gdf = gdf[gdf["cluster"] == n]
                    c_poly = gdf2poly(c_gdf)
                    clusters.append(c_poly)
            else:
                raise FatalError("Clustering support not enabled; unable to import sklearn package")

    # update timing profiles
    profiles[toregion.__name__] = time.perf_counter() - tstart

    # return region
    return {
        "gdf": gdf,
        "poly": polygon, # convex hull of polygons
        "clusters": clusters, # list of polygon clusters for cmr request
        "raster": {
            "geojson": datafile, # geojson file
            "length": len(datafile), # geojson file length
            "cellsize": cellsize  # units are in crs/projection
        }
    }



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import json
import warnings
import datetime
import numpy as np
from sliderule import logger, SLIDERULE_EPSG

# imports with warnings if not present
try:
    import scipy.io
except ModuleNotFoundError as e:
    sys.stderr.write("Warning: missing packages, some functions will throw an exception if called. (%s)\n" % (str(e)))
try:
    import geopandas
    from geopandas.io.arrow import _geopandas_to_arrow
    from geopandas._compat import import_optional_dependency
except ModuleNotFoundError as e:
    sys.stderr.write("Warning: missing packages, some functions will throw an exception if called. (%s)\n" % (str(e)))  
try:
    import h5py
except ModuleNotFoundError as e:
    sys.stderr.write("Warning: missing packages, some functions will throw an exception if called. (%s)\n" % (str(e)))

# attributes for ATL06-SR and ATL03 variables
def get_attributes(**kwargs):
    # set default keyword arguments
    kwargs.setdefault('lon_key','longitude')
    kwargs.setdefault('lat_key','latitude')
    lon_key,lat_key = (kwargs['lon_key'],kwargs['lat_key'])
    coordinates = f'{lat_key} {lon_key}'
    attrs = {}
    # file level attributes
    attrs['featureType'] = 'trajectory'
    attrs['title'] = "ATLAS/ICESat-2 SlideRule Height"
    attrs['reference'] = 'https://doi.org/10.5281/zenodo.5484048'
    attrs['date_created'] = datetime.datetime.now().isoformat()
    attrs['geospatial_lat_units'] = "degrees_north"
    attrs['geospatial_lon_units'] = "degrees_east"
    attrs['geospatial_ellipsoid'] = "WGS84"
    attrs['date_type'] = "UTC"
    attrs['time_type'] = "CCSDS UTC-A"
    # segment ID
    attrs['segment_id'] = {}
    attrs['segment_id']['long_name'] = "Along-track segment ID number"
    attrs['segment_id']['coordinates'] = coordinates
    # delta time
    attrs['delta_time'] = {}
    attrs['delta_time']['units'] = "seconds since 2018-01-01"
    attrs['delta_time']['long_name'] = "Elapsed GPS seconds"
    attrs['delta_time']['standard_name'] = "time"
    attrs['delta_time']['calendar'] = "standard"
    attrs['delta_time']['coordinates'] = coordinates
    # latitude
    attrs[lat_key] = {}
    attrs[lat_key]['units'] = "degrees_north"
    attrs[lat_key]['long_name'] = "Latitude"
    attrs[lat_key]['standard_name'] = "latitude"
    attrs[lat_key]['valid_min'] = -90.0
    attrs[lat_key]['valid_max'] = 90.0
    # longitude
    attrs[lon_key] = {}
    attrs[lon_key]['units'] = "degrees_east"
    attrs[lon_key]['long_name'] = "Longitude"
    attrs[lon_key]['standard_name'] = "longitude"
    attrs[lon_key]['valid_min'] = -180.0
    attrs[lon_key]['valid_max'] = 180.0
    # mean height from fit
    attrs['h_mean'] = {}
    attrs['h_mean']['units'] = "meters"
    attrs['h_mean']['long_name'] = "Height Mean"
    attrs['h_mean']['coordinates'] = coordinates
    # uncertainty in mean height
    attrs['h_sigma'] = {}
    attrs['h_sigma']['units'] = "meters"
    attrs['h_sigma']['long_name'] = "Height Error"
    attrs['h_sigma']['coordinates'] = coordinates
    # RMS of fit
    attrs['rms_misfit'] = {}
    attrs['rms_misfit']['units'] = "meters"
    attrs['rms_misfit']['long_name'] = "RMS of fit"
    attrs['rms_misfit']['coordinates'] = coordinates
    # along track slope
    attrs['dh_fit_dx'] = {}
    attrs['dh_fit_dx']['units'] = "meters/meters"
    attrs['dh_fit_dx']['long_name'] = "Along Track Slope"
    attrs['dh_fit_dx']['coordinates'] = coordinates
    # across track slope
    attrs['dh_fit_dy'] = {}
    attrs['dh_fit_dy']['units'] = "meters/meters"
    attrs['dh_fit_dy']['long_name'] = "Across Track Slope"
    attrs['dh_fit_dy']['coordinates'] = coordinates
    # number of photons in fit
    attrs['n_fit_photons'] = {}
    attrs['n_fit_photons']['units'] = "1"
    attrs['n_fit_photons']['long_name'] = "Number of Photons in Fit"
    attrs['n_fit_photons']['coordinates'] = coordinates
    # surface fit window
    attrs['w_surface_window_final'] = {}
    attrs['w_surface_window_final']['units'] = "meters"
    attrs['w_surface_window_final']['long_name'] = "Surface Window Width"
    attrs['w_surface_window_final']['coordinates'] = coordinates
    # robust dispersion estimate of fit
    attrs['h_robust_sprd'] = {}
    attrs['h_robust_sprd']['units'] = "meters"
    attrs['h_robust_sprd']['long_name'] = "Robust Spread"
    attrs['h_robust_sprd']['coordinates'] = coordinates
    # orbital cycle
    attrs['cycle'] = {}
    attrs['cycle']['long_name'] = "Orbital cycle"
    attrs['cycle']['coordinates'] = coordinates
    # RGT
    attrs['rgt'] = {}
    attrs['rgt']['long_name'] = "Reference Ground Track"
    attrs['rgt']['valid_min'] = 1
    attrs['rgt']['valid_max'] = 1387
    attrs['rgt']['coordinates'] = coordinates
    # ground track
    attrs['gt'] = {}
    attrs['gt']['long_name'] = "Ground track identifier"
    attrs['gt']['flag_values'] = [10, 20, 30, 40, 50, 60]
    attrs['gt']['flag_meanings'] = "GT1L, GT1R, GT2L, GT2R, GT3L, GT3R"
    attrs['gt']['valid_min'] = 10
    attrs['gt']['valid_max'] = 60
    attrs['gt']['coordinates'] = coordinates
    # along-track distance
    attrs['x_atc'] = {}
    attrs['x_atc']['units'] = "meters"
    attrs['x_atc']['long_name'] = "Along track distance from equator"
    attrs['x_atc']['coordinates'] = coordinates
    # across-track distance
    attrs['y_atc'] = {}
    attrs['y_atc']['units'] = "meters"
    attrs['y_atc']['long_name'] = "Across track distance from reference ground track"
    attrs['y_atc']['coordinates'] = coordinates
    # spot
    attrs['spot'] = {}
    attrs['spot']['long_name'] = "ATLAS spot number"
    attrs['spot']['valid_min'] = 1
    attrs['spot']['valid_max'] = 6
    attrs['spot']['coordinates'] = coordinates
    # pflags
    attrs['pflags'] = {}
    attrs['pflags']['long_name'] = "Processing Flags"
    attrs['pflags']['flag_values'] = [0, 1, 2, 4]
    attrs['pflags']['flag_meanings'] = ("valid, spread too short, "
        "too few photons, max iterations reached")
    attrs['pflags']['valid_min'] = 0
    attrs['pflags']['valid_max'] = 4
    attrs['pflags']['coordinates'] = coordinates
    # ATL03 specific variables
    # segment_dist
    attrs['segment_dist'] = {}
    attrs['segment_dist']['units'] = 'meters'
    attrs['segment_dist']['long_name'] = ("Along track distance "
        "from equator for segment")
    attrs['segment_dist']['coordinates'] = coordinates
    # sc_orient
    attrs['sc_orient'] = {}
    attrs['sc_orient']['long_name'] = "Spacecraft Orientation"
    attrs['sc_orient']['flag_values'] = [0, 1, 2]
    attrs['sc_orient']['flag_meanings'] = "backward forward transition"
    attrs['sc_orient']['valid_min'] = 0
    attrs['sc_orient']['valid_max'] = 2
    attrs['sc_orient']['coordinates'] = coordinates
    # track
    attrs['track'] = {}
    attrs['track']['long_name'] = "Pair track identifier"
    attrs['track']['flag_values'] = [1, 2, 3]
    attrs['track']['flag_meanings'] = "PT1, PT2, PT3"
    attrs['track']['valid_min'] = 1
    attrs['track']['valid_max'] = 3
    attrs['track']['coordinates'] = coordinates
    # pair
    attrs['pair'] = {}
    attrs['pair']['long_name'] = "left-right identifier of pair track"
    attrs['pair']['flag_values'] = [0, 1]
    attrs['pair']['flag_meanings'] = "left, right"
    attrs['pair']['valid_min'] = 1
    attrs['pair']['valid_max'] = 3
    attrs['pair']['coordinates'] = coordinates
    # photon height
    attrs['height'] = {}
    attrs['height']['units'] = "meters"
    attrs['height']['long_name'] = "Photon Height"
    attrs['height']['coordinates'] = coordinates
    # photon height
    attrs['height'] = {}
    attrs['height']['units'] = "meters"
    attrs['height']['long_name'] = "Photon Height"
    attrs['height']['coordinates'] = coordinates
    # quality_ph
    attrs['quality_ph'] = {}
    attrs['quality_ph']['long_name'] = "Photon Quality"
    attrs['quality_ph']['flag_values'] = [0, 1, 2, 3]
    attrs['quality_ph']['flag_meanings'] = ("nominal possible_afterpulse "
        "possible_impulse_response_effect possible_tep")
    attrs['quality_ph']['valid_min'] = 1
    attrs['quality_ph']['valid_max'] = 3
    attrs['quality_ph']['coordinates'] = coordinates
    # atl03_cnf
    attrs['atl03_cnf'] = {}
    attrs['atl03_cnf']['long_name'] = "Photon Signal Confidence"
    attrs['atl03_cnf']['flag_values'] = [-2, 1, 0, 1, 2, 3, 4]
    attrs['atl03_cnf']['flag_meanings'] = ("possible_tep "
        "not_considered noise buffer low medium high")
    attrs['atl03_cnf']['valid_min'] = -2
    attrs['atl03_cnf']['valid_max'] = 3
    attrs['atl03_cnf']['coordinates'] = coordinates
    # atl08_class
    attrs['atl08_class'] = {}
    attrs['atl08_class']['long_name'] = "ATL08 Photon Classification"
    attrs['atl08_class']['flag_values'] = [0, 1, 2, 3, 4]
    attrs['atl08_class']['flag_meanings'] = ("noise "
        "ground canopy top_of_canopy unclassified")
    attrs['atl08_class']['valid_min'] = 0
    attrs['atl08_class']['valid_max'] = 4
    attrs['atl08_class']['coordinates'] = coordinates
    # yapc_score
    attrs['yapc_score'] = {}
    attrs['yapc_score']['units'] = "1"
    attrs['yapc_score']['long_name'] = "YAPC Photon Weight"
    attrs['yapc_score']['valid_min'] = 0
    attrs['yapc_score']['valid_max'] = 255
    attrs['yapc_score']['coordinates'] = coordinates
    # return the attributes for the sliderule variables
    return attrs

# PURPOSE: encoder for creating the file attributes
def attributes_encoder(attr):
    """Custom encoder for creating file attributes in Python 3"""
    if isinstance(attr, (bytes, bytearray)):
        return attr.decode('utf-8')
    if isinstance(attr, (np.int_, np.intc, np.intp, np.int8, np.int16, np.int32,
        np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
        return int(attr)
    elif isinstance(attr, (np.float_, np.float16, np.float32, np.float64)):
        return float(attr)
    elif isinstance(attr, (np.ndarray)):
        return attr.tolist()
    elif isinstance(attr, (np.bool_)):
        return bool(attr)
    elif isinstance(attr, (np.void)):
        return None
    else:
        return attr

# conversion factors between time units and seconds
_to_sec = {'microseconds': 1e-6, 'microsecond': 1e-6,
           'microsec': 1e-6, 'microsecs': 1e-6,
           'milliseconds': 1e-3, 'millisecond': 1e-3,
           'millisec': 1e-3, 'millisecs': 1e-3,
           'msec': 1e-3, 'msecs': 1e-3, 'ms': 1e-3,
           'seconds': 1.0, 'second': 1.0, 'sec': 1.0,
           'secs': 1.0, 's': 1.0,
           'minutes': 60.0, 'minute': 60.0,
           'min': 60.0, 'mins': 60.0,
           'hours': 3600.0, 'hour': 3600.0,
           'hr': 3600.0, 'hrs': 3600.0, 'h': 3600.0,
           'day': 86400.0, 'days': 86400.0, 'd': 86400.0}

# PURPOSE: split a date string into units and epoch
def split_date_string(date_string: str):
    """
    split a date string into units and epoch

    Parameters
    ----------
    date_string: str
        time-units since yyyy-mm-dd hh:mm:ss
    """
    try:
        units,_,epoch = date_string.split(None, 2)
    except ValueError:
        raise ValueError(f'Invalid format: {date_string}')
    else:
        return (units.lower(), epoch)

# PURPOSE: convert a numpy datetime array to delta times from the UNIX epoch
def convert_datetime(date, epoch='2018-01-01T00:00:00'):
    """
    Convert a numpy datetime array to seconds since ``epoch``

    Parameters
    ----------
    date: obj
        numpy datetime array
    epoch: str, default '2018-01-01T00:00:00'
        epoch for output delta_time

    Returns
    -------
    delta_time: float
        seconds since epoch
    """
    return (date - np.datetime64(epoch)) / np.timedelta64(1, 's')

# calculate centroid of polygon
def centroid(x,y):
    npts = len(x)
    area,cx,cy = (0.0,0.0,0.0)
    for i in range(npts-1):
        SA = x[i]*y[i+1] - x[i+1]*y[i]
        area += SA
        cx += (x[i] + x[i+1])*SA
        cy += (y[i] + y[i+1])*SA
    cx /= 3.0*area
    cy /= 3.0*area
    return (cx,cy)

# determine if polygon winding is counter-clockwise
def winding(x,y):
    npts = len(x)
    wind = np.sum([(x[i+1] - x[i])*(y[i+1] + y[i]) for i in range(npts - 1)])
    return wind

# geodesic area of a polygon in square kilometers
def area(lon, lat, a=6378.1370):
    assert len(lon) == len(lat), "Longitude and Latitude arrays must be the same length"
    assert len(lon) > 2, "Need at least 3 points to calculate area"
    # factor for converting to radians
    dtr = np.pi/180.0
    # initialize area
    area = 0.0
    npts = len(lon)
    # for each point in the polygon
    for i in range(npts):
        # wrap coordinates
        ii = (i + 1) % npts
        # calculate geodesic area in square kilometers
        area += (a**2)*(lon[ii]*dtr - lon[i]*dtr) * \
            (2.0 + np.sin(lat[i]*dtr) + np.sin(lat[ii]*dtr))/2.0
    # return absolute value of area
    return np.abs(area).squeeze()

# fix longitudes to be -180:180
def wrap_longitudes(lon):
    phi = np.arctan2(np.sin(lon*np.pi/180.0),np.cos(lon*np.pi/180.0))
    # convert phi from radians to degrees
    return phi*180.0/np.pi

# convert coordinates to a sliderule region
def to_region(lon,lat):
    region = [{'lon':ln,'lat':lt} for ln,lt in np.c_[lon,lat]]
    return region

# extract coordinates from a sliderule region
def from_region(polygon):
    npts = len(polygon)
    x = np.zeros((npts))
    y = np.zeros((npts))
    for i,p in enumerate(polygon):
        x[i] = p['lon']
        y[i] = p['lat']
    return (x,y)

# convert geodataframe vector object to a list of sliderule regions
def from_geodataframe(gdf):
    # verify that geodataframe is in latitude/longitude
    geodataframe = gdf.to_crs(SLIDERULE_EPSG)
    # create a list of regions
    regions = []
    # for each region
    for geometry in geodataframe.geometry:
        regions.append([{'lon':ln,'lat':lt} for ln,lt in geometry.exterior.coords])
    # return the list of regions
    return regions

# output request parameters to JSON
def to_json(filename, **kwargs):
    # set default keyword arguments
    kwargs.setdefault('parameters',None)
    kwargs.setdefault('regions',[])
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    # import optional dependencies
    pyproj = import_optional_dependency(
        "pyproj", extra="pyproj is required for CRS conversion."
    )
    # add each parameter as an attribute
    SRparams = ['H_min_win', 'atl08_class', 'atl03_quality', 'ats', 'cnf',
        'cnt', 'len', 'maxi', 'res', 'sigma_r_max', 'srt', 'yapc']
    output = {}
    # for each adjustable sliderule parameter
    for p in SRparams:
        # try to convert the parameter if available
        try:
            output[p] = attributes_encoder(kwargs['parameters'][p])
        except:
            pass
    # save CRS to JSON
    crs = pyproj.CRS.from_string(kwargs['crs'])
    output['crs'] = crs.to_string()
    # save each region following GeoJSON specification
    output['type'] = 'FeatureCollection'
    output['features'] = []
    for i,poly in enumerate(kwargs['regions']):
        lon, lat = from_region(poly)
        lon = attributes_encoder(lon)
        lat = attributes_encoder(lat)
        geometry=dict(type="polygon", coordinates=[])
        geometry['coordinates'].append([[ln,lt] for ln,lt in zip(lon,lat)])
        output['features'].append(dict(type="Feature", geometry=geometry))
    # dump the attributes to a JSON file
    with open(filename, 'w') as fid:
        json.dump(output, fid)
    # print the filename and dictionary structure
    logger.info(filename)
    logger.info(list(output.keys()))

# read request parameters and regions from JSON
def from_json(filename, **kwargs):
    # load the JSON file
    with open(filename, 'r') as fid:
        attributes = json.load(fid)
    # print the filename and dictionary structure
    logger.info(filename)
    logger.info(list(attributes.keys()))
    # try to get the sliderule adjustable parameters
    SRparams = ['H_min_win', 'atl08_class', 'atl03_quality', 'ats', 'cnf',
        'cnt', 'len', 'maxi', 'res', 'sigma_r_max', 'srt', 'yapc']
    parms = {}
    # for each adjustable sliderule parameter
    for p in SRparams:
        # read the parameter if available
        try:
            parms[p] = attributes[p]
        except:
            pass
    # create a list of regions
    regions = []
    # for each feature in the JSON file
    for feature in attributes['features']:
        # for each coordinate set in the feature
        for coords in feature['geometry']['coordinates']:
            # append to sliderule regions
            regions.append([{'lon':ln,'lat':lt} for ln,lt in coords])
    # return the sliderule parameters and regions
    return (parms, regions)

def to_parquet(gdf, filename, **kwargs):
    # set default keyword arguments
    kwargs.setdefault('index',None)
    kwargs.setdefault('compression','snappy')
    kwargs.setdefault('schema_version',None)
    kwargs.setdefault('parameters',dict())
    kwargs.setdefault('regions',[])
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    # import optional dependencies
    pyproj = import_optional_dependency(
        "pyproj", extra="pyproj is required for CRS conversion."
    )
    parquet = import_optional_dependency(
        "pyarrow.parquet", extra="pyarrow is required for Parquet support."
    )
    # output metadata
    output = {}
    # for each adjustable sliderule parameter
    [kwargs['parameters'].pop(p) for p in ['poly']]
    for p,val in kwargs['parameters'].items():
        # try to convert the parameter if available
        try:
            output[p] = attributes_encoder(val)
        except:
            pass
    # save CRS to JSON
    crs = pyproj.CRS.from_string(kwargs['crs'])
    output['crs'] = crs.to_string()
    # save each region following GeoJSON specification
    output['type'] = 'FeatureCollection'
    output['features'] = []
    for i,poly in enumerate(kwargs['regions']):
        lon, lat = from_region(poly)
        lon = attributes_encoder(lon)
        lat = attributes_encoder(lat)
        geometry=dict(type="polygon", coordinates=[])
        geometry['coordinates'].append([[ln,lt] for ln,lt in zip(lon,lat)])
        output['features'].append(dict(type="Feature", geometry=geometry))
    # dump the attributes to encoded JSON-format
    sliderule_metadata = json.dumps(output).encode('utf-8') 
    # convert geodataframe to arrow table
    table = _geopandas_to_arrow(gdf,
        index=kwargs['index'],
        schema_version=kwargs['schema_version']
    )
    # store sliderule specific file-level metadata
    metadata = table.schema.metadata
    metadata.update({b"sliderule": sliderule_metadata})
    # replace schema metadata with updated
    table = table.replace_schema_metadata(metadata)
    # write arrow table to parquet file
    parquet.write_table(table, filename,
        compression=kwargs['compression']
    )

def from_parquet(filename, **kwargs):
    # set default keyword arguments
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    kwargs.setdefault('return_parameters',False)
    kwargs.setdefault('return_regions',False)
    # import optional dependencies
    parquet = import_optional_dependency(
        "pyarrow.parquet", extra="pyarrow is required for Parquet support."
    )
    # read arrow table from parquet file
    # and convert to coordinate reference system
    gdf = geopandas.read_parquet(filename).to_crs(kwargs['crs'])
    # if not returning the query parameters or polygon
    if not (kwargs['return_parameters'] or kwargs['return_regions']):
        # return geodataframe
        return gdf
    # create tuple with returns
    output = (gdf,)
    # get parquet file metadata
    metadata = parquet.read_metadata(filename).metadata
    # validate sliderule metadata
    if b'sliderule' not in metadata.keys():
        logger.error("No sliderule metadata found in Parquet file")
        return output
    # decode sliderule metadata from JSON
    parms = json.loads(metadata[b'sliderule'].decode('utf-8'))
    # create a list of regions
    regions = []
    # for each feature in the JSON file
    for feature in parms.pop('features'):
        # for each coordinate set in the feature
        for coords in feature['geometry']['coordinates']:
            # append to sliderule regions
            regions.append([{'lon':ln,'lat':lt} for ln,lt in coords])
    # if returning the parameters
    if kwargs['return_parameters']:
        # add parameters to output tuple
        [parms.pop(v) for v in ('version','commit','type')]
        output += (parms,)
    # if returning the regions
    if kwargs['return_regions']:
        # add regions to output tuple
        output += (regions,)
    # return the combined tuple
    return output

# output geodataframe to netCDF (version 3)
def to_nc(gdf, filename, **kwargs):
    # add warning that function is deprecated
    logger.critical(f"Deprecated. Will be removed in a future release")
    # set default keyword arguments
    kwargs.setdefault('parameters',None)
    kwargs.setdefault('regions',[])
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    kwargs.setdefault('lon_key','longitude')
    kwargs.setdefault('lat_key','latitude')
    kwargs.setdefault('units','seconds since 2018-01-01T00:00:00')
    # get output attributes
    attributes = get_attributes()
    # open netCDF3 file object (64-bit offset format)
    fileID = scipy.io.netcdf.netcdf_file(filename, 'w', version=2)
    warnings.filterwarnings("ignore")
    # convert geodataframe to pandas dataframe
    df = geopandas.pd.DataFrame(gdf.drop(columns='geometry'))
    # append latitude and longitude as columns
    lon_key,lat_key = (kwargs['lon_key'],kwargs['lat_key'])
    df[lat_key] = gdf['geometry'].values.y
    df[lon_key] = gdf['geometry'].values.x
    # get geodataframe coordinate system
    if gdf.crs:
        kwargs['crs'] = gdf.crs
    # create output delta time dimension and variable
    fileID.createDimension('delta_time', len(df.index))
    nc = fileID.createVariable('delta_time', 'f8', ('delta_time',))
    units, epoch = split_date_string(kwargs['units'])
    delta_time = convert_datetime(df.index, epoch)/_to_sec[units]
    nc[:] = delta_time.copy()
    # set attributes for delta time
    attributes['delta_time']['units'] = kwargs['units']
    attributes['delta_time']['long_name'] = "Elapsed GPS time"
    for att_key,att_val in attributes['delta_time'].items():
        setattr(nc, att_key, att_val)
    # for each variable in the dataframe
    for key,val in df.items():
        if np.issubdtype(val, np.unsignedinteger):
            nc = fileID.createVariable(key, 'i4', ('delta_time',))
            nc[:] = val.astype(np.int32)
        else:
            nc = fileID.createVariable(key, val.dtype, ('delta_time',))
            nc[:] = val.copy()
        # set attributes for variable
        for att_key,att_val in attributes[key].items():
            setattr(nc,att_key,att_val)
    # add file attributes
    fileID.featureType = attributes['featureType']
    fileID.title = attributes['title']
    fileID.reference = attributes['reference']
    fileID.date_created = attributes['date_created']
    fileID.date_type = attributes['date_type']
    fileID.time_type = attributes['time_type']
    # save geodataframe coordinate system
    fileID.crs = kwargs['crs']
    # add geospatial attributes
    if kwargs['crs'] in ('EPSG:4326', SLIDERULE_EPSG):
        fileID.geospatial_lat_units = \
            attributes['geospatial_lat_units']
        fileID.geospatial_lon_units = \
            attributes['geospatial_lon_units']
        fileID.geospatial_ellipsoid = \
            attributes['geospatial_ellipsoid']
    # add each parameter as an attribute
    SRparams = ['H_min_win', 'atl08_class', 'atl03_quality', 'ats', 'cnf',
        'cnt', 'len', 'maxi', 'res', 'sigma_r_max', 'srt', 'yapc']
    # for each adjustable sliderule parameter
    for p in SRparams:
        # try to get the parameter if available
        try:
            attr = attributes_encoder(kwargs['parameters'][p])
            setattr(fileID, p, json.dumps(attr))
        except:
            # if empty or unavailable
            pass
    # for each version parameter
    for p in ['version', 'commit']:
        # try to get the parameter if available
        try:
            setattr(fileID, p, kwargs['parameters'][p])
        except:
            # if empty or unavailable
            pass
    # save each region as a list attribute
    for i,poly in enumerate(kwargs['regions']):
        lon, lat = from_region(poly)
        lon = attributes_encoder(lon)
        lat = attributes_encoder(lat)
        setattr(fileID, 'poly{0:d}_x'.format(i), json.dumps(lon))
        setattr(fileID, 'poly{0:d}_y'.format(i), json.dumps(lat))
    # Output netCDF structure information
    logger.info(filename)
    logger.info(list(fileID.variables.keys()))
    # Closing the netCDF file
    fileID.close()
    warnings.filterwarnings("default")

# input geodataframe from netCDF (version 3)
def from_nc(filename, **kwargs):
    # add warning that function is deprecated
    logger.critical(f"Deprecated. Will be removed in a future release")
    # set default crs
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    kwargs.setdefault('lon_key','longitude')
    kwargs.setdefault('lat_key','latitude')
    kwargs.setdefault('index_key','time')
    kwargs.setdefault('return_parameters',False)
    kwargs.setdefault('return_regions',False)
    # open netCDF3 file object (64-bit offset format)
    fileID = scipy.io.netcdf.netcdf_file(filename, 'r', version=2)
    warnings.filterwarnings("ignore")
    # input dictionary for input variables
    nc = {}
    # get each variable from netCDF
    for key,val in fileID.variables.items():
        # swap byte order to little endian if big endian
        flattened = val[:].squeeze()
        if (flattened.dtype.byteorder == '>'):
            nc[key] = flattened.byteswap().newbyteorder()
        else:
            nc[key] = flattened.copy()
    # get geodataframe coordinate system
    if getattr(fileID, 'crs'):
        kwargs['crs'] = fileID.crs.decode('utf-8')
    # parameter attributes to read
    SRparams = ['H_min_win', 'atl08_class', 'atl03_quality', 'ats', 'cnf',
        'cnt', 'len', 'maxi', 'res', 'sigma_r_max', 'srt', 'yapc']
    # for each adjustable sliderule parameter
    parms = {}
    for p in SRparams:
        # try to get the parameter if available
        try:
            parms[p] = json.loads(getattr(fileID, p))
        except:
            # if empty or unavailable
            pass
    # read each region from list attribute
    regions = []
    # counter variable for reading polygon attributes
    i = 0
    while True:
        # attempt to get x and y coordinates for query polygon
        try:
            x = json.loads(getattr(fileID, f'poly{i:d}_x'))
            y = json.loads(getattr(fileID, f'poly{i:d}_y'))
        except:
            break
        else:
            # convert x and y coordinates into sliderule region
            regions.append(to_region(x, y))
            # add to polygon counter
            i += 1
    # Generate Time Column
    units, epoch = split_date_string(fileID.variables['delta_time'].units.decode('utf-8'))
    delta_time = (1e9*_to_sec[units]*nc['delta_time']).astype('timedelta64[ns]')
    nc['time'] = geopandas.pd.to_datetime(np.datetime64(epoch) + delta_time)
    # remove delta time from dictionary
    del nc['delta_time']
    # generate geometry column
    lon_key,lat_key = (kwargs['lon_key'],kwargs['lat_key'])
    geometry = geopandas.points_from_xy(nc[lon_key],nc[lat_key])
    # remove coordinates from dictionary
    del nc[lon_key]
    del nc[lat_key]
    # Closing the netCDF file
    fileID.close()
    warnings.filterwarnings("default")
    # create Pandas DataFrame object
    df = geopandas.pd.DataFrame(nc)
    # build GeoDataFrame
    gdf = geopandas.GeoDataFrame(df, geometry=geometry, crs=kwargs['crs'])
    # set index
    gdf.set_index(kwargs['index_key'], inplace=True)
    gdf.sort_index(inplace=True)
    # if not returning the query parameters or polygon
    if not (kwargs['return_parameters'] or kwargs['return_regions']):
        # return geodataframe
        return gdf
    # create tuple with returns
    output = (gdf,)
    # if returning the parameters
    if kwargs['return_parameters']:
        # add parameters to output tuple
        output += (parms,)
    # if returning the regions
    if kwargs['return_regions']:
        # add regions to output tuple
        output += (regions,)
    # return the combined tuple
    return output

# output geodataframe to HDF5
def to_hdf(gdf, filename, **kwargs):
    # set default keyword arguments
    kwargs.setdefault('driver','pytables')
    kwargs.setdefault('parameters',None)
    kwargs.setdefault('regions',[])
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    kwargs.setdefault('lon_key','longitude')
    kwargs.setdefault('lat_key','latitude')
    kwargs.setdefault('units','seconds since 2018-01-01T00:00:00')
    # get output attributes
    attributes = get_attributes()
    # convert geodataframe to pandas dataframe
    df = geopandas.pd.DataFrame(gdf.drop(columns='geometry'))
    # append latitude and longitude as columns
    lon_key,lat_key = (kwargs['lon_key'],kwargs['lat_key'])
    df[lat_key] = gdf['geometry'].values.y
    df[lon_key] = gdf['geometry'].values.x
    # get geodataframe coordinate system
    if gdf.crs:
        kwargs['crs'] = str(gdf.crs)
    # output to HDF5 format
    if (kwargs['driver'].lower() == 'pytables'):
        kwargs.pop('driver')
        # write dataframe to pytables HDF5
        write_pytables(df, filename, attributes, **kwargs)
    elif (kwargs['driver'].lower() == 'h5py'):
        kwargs.pop('driver')
        # write dataframe to HDF5
        write_h5py(df, filename, attributes, **kwargs)

# write pandas dataframe to pytables HDF5
def write_pytables(df, filename, attributes, **kwargs):
    # add warning that function is deprecated
    logger.critical(f"Deprecated. Will be removed in a future release")
    # set default keyword arguments
    kwargs.setdefault('parameters',None)
    kwargs.setdefault('regions',[])
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    # write data to a pytables HDF5 file
    df.to_hdf(filename, 'sliderule_segments', format="table", mode="w")
    # add file attributes
    fileID = geopandas.pd.HDFStore(filename, mode='a')
    fileID.root._v_attrs.TITLE = attributes['title']
    fileID.root._v_attrs.reference = attributes['reference']
    fileID.root._v_attrs.date_created = attributes['date_created']
    fileID.root._v_attrs.date_type = attributes['date_type']
    fileID.root._v_attrs.time_type = attributes['time_type']
    # set coordinate reference system as attribute
    fileID.root._v_attrs.crs = kwargs['crs']
    # add geospatial attributes
    if kwargs['crs'] in ('EPSG:4326', SLIDERULE_EPSG):
        fileID.root._v_attrs.geospatial_lat_units = \
            attributes['geospatial_lat_units']
        fileID.root._v_attrs.geospatial_lon_units = \
            attributes['geospatial_lon_units']
        fileID.root._v_attrs.geospatial_ellipsoid = \
            attributes['geospatial_ellipsoid']
    # add each parameter as an attribute
    SRparams = ['H_min_win', 'atl08_class', 'atl03_quality', 'ats', 'cnf',
        'cnt', 'len', 'maxi', 'res', 'sigma_r_max', 'srt', 'yapc']
    # for each adjustable sliderule parameter
    for p in SRparams:
        # try to get the parameter if available
        try:
            attr = attributes_encoder(kwargs['parameters'][p])
            setattr(fileID.root._v_attrs, p, json.dumps(attr))
        except:
            # if empty or unavailable
            pass
    # for each version parameter
    for p in ['version', 'commit']:
        # try to get the parameter if available
        try:
            setattr(fileID.root._v_attrs, p, kwargs['parameters'][p])
        except:
            # if empty or unavailable
            pass
    # save each region as a list attribute
    for i,poly in enumerate(kwargs['regions']):
        lon, lat = from_region(poly)
        lon = attributes_encoder(lon)
        lat = attributes_encoder(lat)
        setattr(fileID.root._v_attrs, f'poly{i:d}_x', json.dumps(lon))
        setattr(fileID.root._v_attrs, f'poly{i:d}_y', json.dumps(lat))
    # Output HDF5 structure information
    logger.info(filename)
    logger.info(fileID.get_storer('sliderule_segments').non_index_axes[0][1])
    # Closing the HDF5 file
    fileID.close()

# write pandas dataframe to h5py HDF5
def write_h5py(df, filename, attributes, **kwargs):
    # add warning that function is deprecated
    logger.critical(f"Deprecated. Will be removed in a future release")
    # set default keyword arguments
    kwargs.setdefault('parameters',None)
    kwargs.setdefault('regions',[])
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    kwargs.setdefault('units','seconds since 2018-01-01T00:00:00')
    # open HDF5 file object
    fileID = h5py.File(filename, mode='w')
    # create HDF5 records
    h5 = {}
    # create output delta time variable
    key = 'delta_time'
    units, epoch = split_date_string(kwargs['units'])
    delta_time = convert_datetime(df.index, epoch)/_to_sec[units]
    h5[key] = fileID.create_dataset(key, delta_time.shape, data=delta_time,
        dtype=delta_time.dtype, compression='gzip')
    # set attributes for delta time
    attributes['delta_time']['units'] = kwargs['units']
    attributes['delta_time']['long_name'] = "Elapsed GPS time"
    for att_key,att_val in attributes[key].items():
        h5[key].attrs[att_key] = att_val
    # for each variable in the dataframe
    for key,val in df.items():
        # create dataset for variable
        h5[key] = fileID.create_dataset(key, val.shape, data=val,
            dtype=val.dtype, compression='gzip')
        h5[key].dims[0].attach_scale(h5['delta_time'])
        # set attributes for variable
        for att_key,att_val in attributes[key].items():
            h5[key].attrs[att_key] = att_val
    # add file attributes
    fileID.attrs['featureType'] = attributes['featureType']
    fileID.attrs['title'] = attributes['title']
    fileID.attrs['reference'] = attributes['reference']
    fileID.attrs['date_created'] = attributes['date_created']
    fileID.attrs['date_type'] = attributes['date_type']
    fileID.attrs['time_type'] = attributes['time_type']
    # set coordinate reference system as attribute
    fileID.attrs['crs'] = kwargs['crs']
    # add geospatial attributes
    if kwargs['crs'] in ('EPSG:4326', SLIDERULE_EPSG):
        fileID.attrs['geospatial_lat_units'] = \
            attributes['geospatial_lat_units']
        fileID.attrs['geospatial_lon_units'] = \
            attributes['geospatial_lon_units']
        fileID.attrs['geospatial_ellipsoid'] = \
            attributes['geospatial_ellipsoid']
    # add each parameter as an attribute
    SRparams = ['H_min_win', 'atl08_class', 'atl03_quality', 'ats', 'cnf',
        'cnt', 'len', 'maxi', 'res', 'sigma_r_max', 'srt', 'yapc']
    # for each adjustable sliderule parameter
    for p in SRparams:
        # try to get the parameter if available
        try:
            attr = attributes_encoder(kwargs['parameters'][p])
            fileID.attrs[p] = json.dumps(attr)
        except:
            # if empty or unavailable
            pass
    # for each version parameter
    for p in ['version', 'commit']:
        # try to get the parameter if available
        try:
            fileID.attrs[p] = kwargs['parameters'][p]
        except:
            # if empty or unavailable
            pass
    # save each region as a list attribute
    for i,poly in enumerate(kwargs['regions']):
        lon, lat = from_region(poly)
        lon = attributes_encoder(lon)
        lat = attributes_encoder(lat)
        fileID.attrs[f'poly{i:d}_x'] = json.dumps(lon)
        fileID.attrs[f'poly{i:d}_y'] = json.dumps(lat)
    # Output HDF5 structure information
    logger.info(filename)
    logger.info(list(fileID.keys()))
    # Closing the HDF5 file
    fileID.close()

# input geodataframe from HDF5
def from_hdf(filename, **kwargs):
    # set default keyword arguments
    kwargs.setdefault('driver','pytables')
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    kwargs.setdefault('lon_key','longitude')
    kwargs.setdefault('lat_key','latitude')
    kwargs.setdefault('return_parameters',False)
    kwargs.setdefault('return_regions',False)
    if (kwargs['driver'].lower() == 'pytables'):
        kwargs.pop('driver')
        # return GeoDataFrame from pytables
        return read_pytables(filename, **kwargs)
    elif (kwargs['driver'].lower() == 'h5py'):
        kwargs.pop('driver')
        # return GeoDataFrame from h5py
        return read_h5py(filename, **kwargs)

# read pandas dataframe from pytables HDF5
def read_pytables(filename, **kwargs):
    # add warning that function is deprecated
    logger.critical(f"Deprecated. Will be removed in a future release")
    # set default crs
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    kwargs.setdefault('lon_key','longitude')
    kwargs.setdefault('lat_key','latitude')
    kwargs.setdefault('return_parameters',False)
    kwargs.setdefault('return_regions',False)
    # open pytables HDF5 to read pandas dataframe
    df = geopandas.pd.read_hdf(filename, **kwargs)
    # generate geometry column
    lon_key,lat_key = (kwargs['lon_key'],kwargs['lat_key'])
    geometry = geopandas.points_from_xy(df[lon_key],df[lat_key])
    # get geodataframe coordinate system from attributes
    fileID = geopandas.pd.HDFStore(filename, mode='r')
    if getattr(fileID.root._v_attrs, 'crs'):
        kwargs['crs'] = str(fileID.root._v_attrs.crs)
    # parameter attributes to read
    SRparams = ['H_min_win', 'atl08_class', 'atl03_quality', 'ats', 'cnf',
        'cnt', 'len', 'maxi', 'res', 'sigma_r_max', 'srt', 'yapc']
    # for each adjustable sliderule parameter
    parms = {}
    for p in SRparams:
        # try to get the parameter if available
        try:
            parms[p] = json.loads(getattr(fileID.root._v_attrs,p))
        except:
            # if empty or unavailable
            pass
    # read each region from list attribute
    regions = []
    # counter variable for reading polygon attributes
    i = 0
    while True:
        # attempt to get x and y coordinates for query polygon
        try:
            x = json.loads(getattr(fileID.root._v_attrs,f'poly{i:d}_x'))
            y = json.loads(getattr(fileID.root._v_attrs,f'poly{i:d}_y'))
        except:
            break
        else:
            # convert x and y coordinates into sliderule region
            regions.append(to_region(x,y))
            # add to polygon counter
            i += 1
    # Closing the HDF5 file
    fileID.close()
    # build and return GeoDataFrame
    gdf = geopandas.GeoDataFrame(df.drop(columns=[lon_key,lat_key]),
        geometry=geometry, crs=kwargs['crs'])
    gdf.sort_index(inplace=True)
    # if not returning the query parameters or polygon
    if not (kwargs['return_parameters'] or kwargs['return_regions']):
        # return geodataframe
        return gdf
    # create tuple with returns
    output = (gdf,)
    # if returning the parameters
    if kwargs['return_parameters']:
        # add parameters to output tuple
        output += (parms,)
    # if returning the regions
    if kwargs['return_regions']:
        # add regions to output tuple
        output += (regions,)
    # return the combined tuple
    return output

# read pandas dataframe from h5py HDF5
def read_h5py(filename, **kwargs):
    # add warning that function is deprecated
    logger.critical(f"Deprecated. Will be removed in a future release")
    # set default crs
    kwargs.setdefault('crs',SLIDERULE_EPSG)
    kwargs.setdefault('lon_key','longitude')
    kwargs.setdefault('lat_key','latitude')
    kwargs.setdefault('index_key','time')
    kwargs.setdefault('return_parameters',False)
    kwargs.setdefault('return_regions',False)
    # open HDF5 file object
    fileID = h5py.File(filename, mode='r')
    # input dictionary for input variables
    h5 = {}
    # get each variable from HDF5
    for key,val in fileID.items():
        h5[key] = val[:].squeeze()
    # get geodataframe coordinate system from attributes
    if 'crs' in fileID.attrs.keys():
        kwargs['crs'] = str(fileID.attrs['crs'])
    # parameter attributes to read
    SRparams = ['H_min_win', 'atl08_class', 'atl03_quality', 'ats', 'cnf',
        'cnt', 'len', 'maxi', 'res', 'sigma_r_max', 'srt', 'yapc']
    # for each adjustable sliderule parameter
    parms = {}
    for p in SRparams:
        # try to get the parameter if available
        try:
            parms[p] = json.loads(fileID.attrs[p])
        except:
            # if empty or unavailable
            pass
    # read each region from list attribute
    regions = []
    # counter variable for reading polygon attributes
    i = 0
    while True:
        # attempt to get x and y coordinates for query polygon
        try:
            x = json.loads(fileID.attrs[f'poly{i:d}_x'])
            y = json.loads(fileID.attrs[f'poly{i:d}_y'])
        except:
            break
        else:
            # convert x and y coordinates into sliderule region
            regions.append(to_region(x,y))
            # add to polygon counter
            i += 1
    # Generate Time Column
    units, epoch = split_date_string(fileID['delta_time'].attrs['units'])
    delta_time = (1e9*_to_sec[units]*h5['delta_time']).astype('timedelta64[ns]')
    h5['time'] = geopandas.pd.to_datetime(np.datetime64(epoch) + delta_time)
    # remove delta time from dictionary
    del h5['delta_time']
    # generate geometry column
    lon_key,lat_key = (kwargs['lon_key'],kwargs['lat_key'])
    geometry = geopandas.points_from_xy(h5[lon_key],h5[lat_key])
    # remove coordinates from dictionary
    del h5[lon_key]
    del h5[lat_key]
    # Closing the HDF5 file
    fileID.close()
    # create Pandas DataFrame object
    df = geopandas.pd.DataFrame(h5)
    # build GeoDataFrame
    gdf = geopandas.GeoDataFrame(df, geometry=geometry, crs=kwargs['crs'])
    # set index
    gdf.set_index(kwargs['index_key'], inplace=True)
    gdf.sort_index(inplace=True)
    # if not returning the query parameters or polygon
    if not (kwargs['return_parameters'] or kwargs['return_regions']):
        # return geodataframe
        return gdf
    # create tuple with returns
    output = (gdf,)
    # if returning the parameters
    if kwargs['return_parameters']:
        # add parameters to output tuple
        output += (parms,)
    # if returning the regions
    if kwargs['return_regions']:
        # add regions to output tuple
        output += (regions,)
    # return the combined tuple
    return output

# output formats wrapper
def to_file(gdf, filename, format='parquet', **kwargs):
    if format.lower() in ('hdf','hdf5','h5'):
        to_hdf(gdf, filename, **kwargs)
    elif format.lower() in ('netcdf','nc'):
        to_nc(gdf, filename, **kwargs)
    elif format.lower() in ('geojson','csv','shp'):
        gdf.to_file(filename, **kwargs)
    elif format.lower() in ('geoparquet','parquet'):
        to_parquet(gdf, filename, **kwargs)

# input formats wrapper
def from_file(filename, format='parquet', **kwargs):
    if format.lower() in ('hdf','hdf5','h5'):
        return from_hdf(filename, **kwargs)
    elif format.lower() in ('netcdf','netcdf4','nc'):
        return from_nc(filename, **kwargs)
    elif format.lower() in ('geojson','csv','shp'):
        return geopandas.from_file(filename, **kwargs)
    elif format.lower() in ('geoparquet','parquet'):
        return from_parquet(filename, **kwargs)
